from xpxchain import models
from tests import harness

BLOCK_VALIDATOR = [
    lambda x: (x.channel_name, 'block'),
    lambda x: (isinstance(x.message, models.BlockInfo), True),
    lambda x: (x.message.height >= 1, True),
]


@harness.listener_test_case({
    'tests': [
        {
            'name': 'test_new_block',
            'subscriptions': ['new_block'],
            'validation': [BLOCK_VALIDATOR],
        },
    ],
})
class TestListener(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config
from tests import responses
from tests import aitertools
import datetime
import binascii
import asyncio
import hashlib
import os
from xpxchain import util
import time
from binascii import hexlify
#import nest_asyncio
#nest_asyncio.apply()

M = 1000000
M1 = M
M10 = 10 * M
M100 = 100 * M
M1000 = 1000 * M

class Error(Exception):
    pass

class TestTransactionHttp(harness.TestCase):

    alice: models.Account
    bob: models.Account
    mike: models.Account

    def __init__(self, task) -> None:
        super().__init__(task)

        if (task == 'test_get_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

        elif (task == 'test_get_transactions'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

        elif (task == 'test_get_transaction_status'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

        elif (task == 'test_get_transaction_statuses'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

        elif (task == 'test_transfer_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

        elif (task == 'test_message_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M10)
        
        elif (task == 'test_account_link_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
        
        elif (task == 'test_modify_account_property_address_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
        
        elif (task == 'test_modify_account_property_mosaic_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
        
        elif (task == 'test_modify_account_property_entity_type_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
        
        elif (task == 'test_register_namespace_transaction'):
            self.mike = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.mike, 4 * M1000)
        
        elif (task == 'test_secret_lock_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
            self.send_funds(config.nemesis, self.bob, M100)
        
        elif (task == 'test_aggregate_transaction_with_cosigners'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
            self.send_funds(config.nemesis, self.bob, M100)

        elif (task == 'test_aggregate_bonded_transaction'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
            self.send_funds(config.nemesis, self.bob, M100)

        elif (task == 'test_create_multisig_and_send_funds'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.multisig = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, M100)
            self.send_funds(config.nemesis, self.bob, M100)
            self.send_funds(config.nemesis, self.multisig, M100)

    
    async def listen(self, account):
        async with client.Listener(f'{responses.ENDPOINT}/ws') as listener:
            await listener.confirmed(account.address)
            await listener.status(account.address)
       
            async for m in listener:
                if (m.channel_name == 'status'):
                    raise Error(m.message)
                elif (m.channel_name == 'confirmedAdded'):
                    return m.message
    
    async def listen_bonded(self, account):
        async with client.Listener(f'{responses.ENDPOINT}/ws') as listener:
            await listener.aggregate_bonded_added(account.address)
            await listener.status(account.address)
       
            async for m in listener:
                if (m.channel_name == 'status'):
                    raise Error(m.message)
                elif (m.channel_name == 'partialAdded'):
                    return m.message

    def send_funds(self, sender, recipient, amount):
        tx = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=recipient.address,
            mosaics=[models.Mosaic(config.mosaic_id, amount)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_with(sender, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(sender)
        self.assertEqual(isinstance(tx, models.TransferTransaction), True)
        self.assertEqual(tx.recipient, recipient.address)
        self.assertEqual(len(tx.mosaics), 1)
        self.assertEqual(tx.mosaics[0].id, config.mosaic_id)
        self.assertEqual(tx.mosaics[0].amount, amount)

        return tx.transaction_info.hash
    

    # TESTS
    def test_get_transaction(self):
        hash = self.send_funds(config.nemesis, self.alice, M10)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            reply = http.get_transaction(hash)
            self.assertEqual(isinstance(reply, models.TransferTransaction), True)
            self.assertEqual(reply.transaction_info.hash, hash)

    
    def test_get_transactions(self):
        hash1 = self.send_funds(config.nemesis, self.alice, M10)
        hash2 = self.send_funds(config.nemesis, self.alice, M10)

        hashes = [hash1, hash2]
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            reply = http.get_transactions(hashes)
            self.assertEqual(len(reply), 2)
            self.assertEqual(isinstance(reply[0], models.TransferTransaction), True)
            self.assertEqual(reply[0].transaction_info.hash in hashes, True)
            self.assertEqual(reply[1].transaction_info.hash in hashes, True)

    
    def test_get_transaction_status(self):
        hash = self.send_funds(config.nemesis, self.alice, M10)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            reply = http.get_transaction_status(hash)
            self.assertEqual(isinstance(reply, models.TransactionStatus), True)
            self.assertEqual(reply.hash, hash)


    def test_get_transaction_statuses(self):
        hash1 = self.send_funds(config.nemesis, self.alice, M10)
        hash2 = self.send_funds(config.nemesis, self.alice, M10)

        hashes = [hash1, hash2]
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            reply = http.get_transaction_statuses(hashes)
            self.assertEqual(len(reply), 2)
            self.assertEqual(isinstance(reply[0], models.TransactionStatus), True)
            self.assertEqual(reply[0].hash in hashes, True)
            self.assertEqual(reply[1].hash in hashes, True)

    
#    def test_transfer_transaction(self): 
#        self.send_funds(config.nemesis, self.alice, M10)


    def test_message_transaction(self):
        message = models.PlainMessage(b'Hello world')

        tx = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=self.bob.address,
            network_type=models.NetworkType.MIJIN_TEST,
            message=message
        )

        signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.TransferTransaction), True)
        self.assertEqual(tx.recipient, self.bob.address)
        self.assertEqual(tx.message, message)

    
    def test_account_link_transaction(self):
        tx = models.AccountLinkTransaction.create(
            deadline=models.Deadline.create(),
            remote_account_key=self.bob.public_key,
            link_action=models.LinkAction.LINK,
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.AccountLinkTransaction), True)
        self.assertEqual(tx.remote_account_key, self.bob.public_key.upper())
        self.assertEqual(tx.link_action, models.LinkAction.LINK)
        
        tx = models.AccountLinkTransaction.create(
            deadline=models.Deadline.create(),
            remote_account_key=self.bob.public_key,
            link_action=models.LinkAction.UNLINK,
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.AccountLinkTransaction), True)
        self.assertEqual(tx.remote_account_key, self.bob.public_key.upper())
        self.assertEqual(tx.link_action, models.LinkAction.UNLINK)
    
    
    def test_modify_account_property_address_transaction(self):
        for property_type in [models.PropertyType.ALLOW_ADDRESS, models.PropertyType.BLOCK_ADDRESS]:
            for modification_type in [models.PropertyModificationType.ADD, models.PropertyModificationType.REMOVE]:
        
                tx = models.ModifyAccountPropertyAddressTransaction.create(
                    deadline=models.Deadline.create(),
                    network_type=models.NetworkType.MIJIN_TEST,
                    property_type=property_type,
                    modifications=[models.AccountPropertyModification(modification_type, self.bob.address)]
                )
                
                signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

                with client.TransactionHTTP(responses.ENDPOINT) as http:
                    http.announce(signed_tx)

                tx = self.listen(self.alice)
                self.assertEqual(isinstance(tx, models.ModifyAccountPropertyAddressTransaction), True)
                self.assertEqual(len(tx.modifications), 1)
                self.assertEqual(tx.property_type, property_type)
                self.assertEqual(tx.modifications[0].modification_type, modification_type)

        
    def test_modify_account_property_mosaic_transaction(self):
        for property_type in [models.PropertyType.ALLOW_MOSAIC, models.PropertyType.BLOCK_MOSAIC]:
            for modification_type in [models.PropertyModificationType.ADD, models.PropertyModificationType.REMOVE]:

                tx = models.ModifyAccountPropertyMosaicTransaction.create(
                    deadline=models.Deadline.create(),
                    network_type=models.NetworkType.MIJIN_TEST,
                    property_type=property_type,
                    modifications=[models.AccountPropertyModification(modification_type, config.mosaic_id)]
                )
                
                signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

                with client.TransactionHTTP(responses.ENDPOINT) as http:
                    http.announce(signed_tx)

                tx = self.listen(self.alice)
                self.assertEqual(isinstance(tx, models.ModifyAccountPropertyMosaicTransaction), True)
                self.assertEqual(len(tx.modifications), 1)
                self.assertEqual(tx.property_type, property_type)
                self.assertEqual(tx.modifications[0].modification_type, modification_type)
                self.assertEqual(tx.modifications[0].value, config.mosaic_id)

        
    def test_modify_account_property_entity_type_transaction(self):
        tx_type = models.TransactionType.AGGREGATE_COMPLETE

        for property_type in [models.PropertyType.BLOCK_TRANSACTION]:
            for modification_type in [models.PropertyModificationType.ADD, models.PropertyModificationType.REMOVE]:

                tx = models.ModifyAccountPropertyEntityTypeTransaction.create(
                    deadline=models.Deadline.create(),
                    network_type=models.NetworkType.MIJIN_TEST,
                    property_type=property_type,
                    modifications=[models.AccountPropertyModification(modification_type, tx_type)]
                )
                
                signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

                with client.TransactionHTTP(responses.ENDPOINT) as http:
                    http.announce(signed_tx)

                tx = self.listen(self.alice)
                self.assertEqual(isinstance(tx, models.ModifyAccountPropertyEntityTypeTransaction), True)
                self.assertEqual(len(tx.modifications), 1)
                self.assertEqual(tx.property_type, property_type)
                self.assertEqual(tx.modifications[0].modification_type, modification_type)
                self.assertEqual(tx.modifications[0].value, tx_type)

    
    def test_register_namespace_transaction(self):
        namespace_name = 'foo' + hexlify(os.urandom(4)).decode('utf-8')
        #namespace_name = 'foo'

        # Create namespace
        tx = models.RegisterNamespaceTransaction.create_root_namespace(
            deadline=models.Deadline.create(),
            network_type=models.NetworkType.MIJIN_TEST,
            namespace_name=namespace_name,
            duration=60
        )
        
        signed_tx = tx.sign_with(self.mike, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.mike)
        self.assertEqual(isinstance(tx, models.RegisterNamespaceTransaction), True)
        self.assertEqual(tx.namespace_name, namespace_name)

        # Create sub namespace
        tx = models.RegisterNamespaceTransaction.create_sub_namespace(
            deadline=models.Deadline.create(),
            network_type=models.NetworkType.MIJIN_TEST,
            namespace_name='bar',
            parent_namespace=namespace_name
        )
        
        signed_tx = tx.sign_with(self.mike, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.mike)
        self.assertEqual(isinstance(tx, models.RegisterNamespaceTransaction), True)
        self.assertEqual(tx.namespace_name, 'bar')
    
        mikes_namespace = namespace_name + ".bar"

        # Link address alias
        for action_type in [models.AliasActionType.LINK, models.AliasActionType.UNLINK]:
            tx = models.AddressAliasTransaction.create(
                deadline=models.Deadline.create(),
                network_type=models.NetworkType.MIJIN_TEST,
                max_fee=1,
                action_type=action_type,
                namespace_id=models.NamespaceId(mikes_namespace),
                address=self.mike.address
            )
            
            signed_tx = tx.sign_with(self.mike, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.mike)
            self.assertEqual(isinstance(tx, models.AddressAliasTransaction), True)
            self.assertEqual(tx.action_type, action_type)
            self.assertEqual(tx.address, self.mike.address)

        # Create mosaic
        nonce = models.MosaicNonce(1)
        mosaic_id = models.MosaicId.create_from_nonce(nonce, self.mike)

        tx = models.MosaicDefinitionTransaction.create(
            deadline=models.Deadline.create(),
            network_type=models.NetworkType.MIJIN_TEST,
            nonce=nonce,
            mosaic_id=mosaic_id,
            mosaic_properties=models.MosaicProperties(0x3, 3),
        )
        
        signed_tx = tx.sign_with(self.mike, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)
        
        tx = self.listen(self.mike)
        self.assertEqual(isinstance(tx, models.MosaicDefinitionTransaction), True)
        self.assertEqual(tx.mosaic_id, mosaic_id)

        # Link mosaic alias
        for action_type in [models.AliasActionType.LINK, models.AliasActionType.UNLINK]:
            tx = models.MosaicAliasTransaction.create(
                deadline=models.Deadline.create(),
                network_type=models.NetworkType.MIJIN_TEST,
                max_fee=1,
                action_type=action_type,
                namespace_id=models.NamespaceId(mikes_namespace),
                mosaic_id=mosaic_id,
            )

            signed_tx = tx.sign_with(self.mike, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.mike)
            self.assertEqual(isinstance(tx, models.MosaicAliasTransaction), True)
            self.assertEqual(tx.mosaic_id, mosaic_id)
            self.assertEqual(tx.action_type, action_type)
   

    def test_secret_lock_transaction(self):
        random_bytes = os.urandom(20)
        h = hashlib.sha3_256(random_bytes)
        secret = binascii.hexlify(h.digest()).decode('utf-8').upper()
        proof = binascii.hexlify(random_bytes).decode('utf-8').upper()

        tx = models.SecretLockTransaction.create(
            deadline=models.Deadline.create(),
            network_type=models.NetworkType.MIJIN_TEST,
            mosaic=models.Mosaic(config.mosaic_id, M1),
            duration=60,
            hash_type=models.HashType.SHA3_256,
            secret=secret,
            recipient=self.bob.address,
        )
        
        signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.SecretLockTransaction), True)
        self.assertEqual(tx.recipient, self.bob.address)
        self.assertEqual(tx.secret, secret)
       
        tx = models.SecretProofTransaction.create(
            deadline=models.Deadline.create(),
            network_type=models.NetworkType.MIJIN_TEST,
            hash_type=models.HashType.SHA3_256,
            secret=secret,
            proof=proof,
            recipient=self.bob.address,
        )
        
        signed_tx = tx.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.SecretProofTransaction), True)
        self.assertEqual(tx.recipient, self.bob.address)
        self.assertEqual(tx.secret, secret)
    

    def test_aggregate_transaction_with_cosigners(self):
        alice_to_bob = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=self.bob.address,
            mosaics=[models.Mosaic(config.mosaic_id, M10)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        bob_to_alice = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=self.alice.address,
            mosaics=[models.Mosaic(config.mosaic_id, M10)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[bob_to_alice.to_aggregate(self.bob), alice_to_bob.to_aggregate(self.alice)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(self.alice, config.gen_hash, [self.bob], fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)
        
        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)
        self.assertEqual(len(tx.inner_transactions), 2)
        self.assertEqual(tx.inner_transactions[0].recipient, self.alice.address)
        self.assertEqual(tx.inner_transactions[1].recipient, self.bob.address)
       
    
    def test_aggregate_bonded_transaction(self):
        alice_to_bob = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=self.bob.address,
            mosaics=[models.Mosaic(config.mosaic_id, M10)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        bob_to_alice = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=self.alice.address,
            mosaics=[models.Mosaic(config.mosaic_id, M10)],
            network_type=models.NetworkType.MIJIN_TEST,
        )
        
        bonded = models.AggregateTransaction.create_bonded(
            deadline=models.Deadline.create(),
            inner_transactions=[bob_to_alice.to_aggregate(self.bob), alice_to_bob.to_aggregate(self.alice)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_bonded = bonded.sign_transaction_with_cosignatories(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        lock = models.LockFundsTransaction.create(
            deadline=models.Deadline.create(),
            network_type=models.NetworkType.MIJIN_TEST,
            mosaic=models.Mosaic(config.mosaic_id, M10),
            duration=60,
            signed_transaction=signed_bonded,
        )
        
        signed_lock = lock.sign_with(self.alice, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_lock)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.LockFundsTransaction), True)
       
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce_partial(signed_bonded)

        tx = self.listen_bonded(self.alice)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        with client.AccountHTTP(responses.ENDPOINT) as http:
            reply = http.aggregate_bonded_transactions(self.bob)
            self.assertEqual(isinstance(reply[0], models.Transaction), True)
            self.assertEqual(isinstance(reply[0], models.AggregateTransaction), True)
            self.assertEqual(isinstance(reply[0], models.AggregateBondedTransaction), True)

            signed_cosig = models.CosignatureTransaction.create(reply[0]).sign_with(self.bob)
            
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce_cosignature(signed_cosig)

        tx = self.listen(self.bob)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)


    def test_create_multisig_and_send_funds(self):
        change_to_multisig = models.ModifyMultisigAccountTransaction.create(
            deadline=models.Deadline.create(),
            min_approval_delta=2,
            min_removal_delta=1,
            modifications=[
                models.MultisigCosignatoryModification.create(self.alice, models.MultisigCosignatoryModificationType.ADD),
                models.MultisigCosignatoryModification.create(self.bob, models.MultisigCosignatoryModificationType.ADD),
            ],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[change_to_multisig.to_aggregate(self.multisig)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(self.multisig, config.gen_hash, [self.alice, self.bob], fee_strategy=util.FeeCalculationStrategy.MEDIUM)
        
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.multisig)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)
        
        self.multisig_to_nemesis = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=config.nemesis.address,
            mosaics=[models.Mosaic(config.mosaic_id, M10)],
            network_type=models.NetworkType.MIJIN_TEST,
        )
            
        tx = models.AggregateTransaction.create_complete(
            deadline=models.Deadline.create(),
            inner_transactions=[self.multisig_to_nemesis.to_aggregate(self.multisig)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_transaction_with_cosignatories(self.alice, config.gen_hash, [self.bob], fee_strategy=util.FeeCalculationStrategy.MEDIUM)
      
        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(self.alice)
        self.assertEqual(isinstance(tx, models.AggregateTransaction), True)
        



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config


@harness.http_test_case({
    'clients': (client.ContractHTTP, client.AsyncContractHTTP),
    'tests': [
    ],
})
class TestContractHttp(harness.TestCase):
    pass



"""
    mosaic_levy_type
    ================

    Constants for the mosaic levy type.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import enum
from ... import util

__all__ = ['MosaicLevyType']


# TODO(ahuszagh) This is not yet implemented in Catapult. Subject to change
# Any commented-out code is identical to the first version of NEM.
@util.inherit_doc
class MosaicLevyType(util.U8Mixin, util.EnumMixin, enum.IntEnum):
    """Mosaic levy type."""

    ABSOLUTE = 1
    CALCULATED = 2

    def description(self) -> str:
        return DESCRIPTION[self]


DESCRIPTION = {
    MosaicLevyType.ABSOLUTE: (
        "The levy is an absolute fee. The field 'fee' states "
        "how many sub-units of the specified mosaic will be "
        "transferred to the recipient."
    ),
    MosaicLevyType.CALCULATED: (
        "The levy is calculated from the transferred amount. "
        "The field 'fee' states how many percentiles of the "
        "transferred quantity will transferred to the recipient."
    ),
}




from __future__ import annotations

import typing

from .mosaic_metadata import MosaicMetadata
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MosaicMetadataInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicMetadataInfo(util.DTO):
    """

    :param metadata:  Mosaic metadata info.

    DTO Format:
        .. code-block:: yaml

            MosaicMetadataInfoDTO:
                metedata: MosaicMetadataDTO
    """

    metadata: MosaicMetadata

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'metadata'}
        
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'metadata': self.metadata.to_dto(),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            metadata=MosaicMetadata.create_from_dto(data['metadata'])
        )



"""
    merkle_proof_info
    =================

    Describes a merkle proof.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .merkle_path_item import MerklePathItem
from .network_type import OptionalNetworkType
from ... import util

import logging
logging.basicConfig(format='[%(filename)s:%(lineno)d] %(levelname)s: %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

__all__ = ['MerkleProofInfo']


# TODO(ahuszagh) Add unittests.
@util.inherit_doc
@util.dataclass(frozen=True)
class MerkleProofInfo(util.DTO):
    """
    Merkle proof information.

    :param merkle_path: The complementary data needed to calculate the merkle root.

    DTO Format:
        .. code-block:: yaml

            MerkleProofInfoDTO:
                merklePath: MerklePathItemDTO[]
    """

    merkle_path: typing.Sequence[MerklePathItem]

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""


        required_l1 = {'merklePath'}
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'merklePath': [i.to_dto(network_type) for i in self.payload],
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        path = data['merklePath']
        return cls(
            merkle_path=[MerklePathItem.create_from_dto(i, network_type) for i in path]
        )



"""
    format
    ======

    Class-bound helpers to simplify conversion between interchange formats.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from .receipt_type import ReceiptType
from .receipt_version import ReceiptVersion
from ..blockchain.network_type import NetworkType
from ... import util

import logging
logging.basicConfig(format='[%(filename)s:%(lineno)d] %(levelname)s: %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

class FormatBase(util.Object):
    """Utilities to simplify loading and saving to interchange formats."""

    def save(self, key, data, value) -> None:
    #def save(self, key, data, value, network_type) -> None:
        """Save value to interchange format by key."""
        raise util.AbstractMethodError

    def load(self, key, data):
    #def load(self, key, data, network_type):
        """Load value from interchange format by key."""
        raise util.AbstractMethodError

    def load_size(self, data):
        """Load transaction size."""
        return self.load('size', data)
        #return self.load('size', data, None)

    def load_type(self, data):
        """Load transaction type."""
        #return self.load('version', data, None)
        return self.load('version', data)

    #def load_network_type(self, data):
    #    """Load network type."""
    #    return self.load('network_type', data, None)

    def find_receipt(self, type_map, data):
        """Find derived receipt class via the receipt type."""
        return type_map[self.load_type(data)]



# DTO HELPERS
@util.dataclass(frozen=True)
class DTOFormat(FormatBase):
    """Utilities to simplify loading and saving to catbuffer."""

    names: typing.Dict[str, str]

    def save(self, key, data, value) -> None:
    #def save(self, key, data, value, network_type) -> None:
        name = self.names[key]
        cb = SAVE_DTO[key]
        if value is not None:
            data[name] = cb(value)
            #data[name] = cb(value, network_type)

    def load(self, key, data):
    #def load(self, key, data, network_type):
        name = self.names[key]
        cb = LOAD_DTO[key]
        value = data.get(name)
        if value is not None:
            return cb(value)
            #return cb(value, network_type)


#def save_version_dto(version, network_type):
#    return version | (int(network_type) << 8)
def save_version_dto(version):
    return version


SAVE_DTO = {
    'version': save_version_dto,
    'type': lambda x, n: x.to_dto(n),
}


#def load_version_dto(data, network_type):
#    return data & 0xFF
def load_version_dto(data):
    return data


#def load_network_type_dto(data, network_type):
#    return NetworkType((data >> 24) & 0x000000ff)


LOAD_DTO = {
    'version': load_version_dto,
    #'network_type': load_network_type_dto,
    'type': ReceiptType.create_from_dto,
}



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config


@harness.http_test_case({
    'clients': (client.BlockchainHTTP, client.AsyncBlockchainHTTP),
    'tests': [
        {
            #/chain/height
            'name': 'test_get_blockchain_height',
            'params': [],
            'method': 'get_blockchain_height',
            'validation': [
                lambda x: (isinstance(x, int), True),
                lambda x: (x >= 1, True),
            ]
        },
        {
            #/chain/score
            'name': 'test_get_blockchain_score',
            'params': [],
            'method': 'get_blockchain_score',
            'validation': [
                lambda x: (isinstance(x, models.BlockchainScore), True),
            ]
        },
    ],
})
class TestChainHttp(harness.TestCase):
    pass



"""
    mosaic_levy
    ===========

    Detailed information describing a levy, or tax on asset transfers.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# type: ignore
# flake8: noqa
# TODO(ahuszagh) This is not yet implemented in Catapult. Subject to change
# Any commented-out code is identical to the first version of NEM.

from __future__ import annotations
import typing

# #from .mosaic_id import MosaicId
# #from .mosaic_levy_type import MosaicLevyType
# #from ..account.address import Address
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['MosaicLevy']


@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicLevy(util.DTO):
    """
    Information describing a mosaic levy.

    :param type: Levy type.
    :param recipient: Recipient of levy.
    :param mosaic_id: Mosaic in which levy is paid.
    :param fee: Fee amount for levy.

    DTO Format:
        .. code-block:: yaml

            MosaicLevyDTO: null
    """

    # #type: MosaicLevyType
    # #recipient: Address
    # #mosaic_id: MosaicId
    # #fee: int

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""
        return data == {}

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {}
        # #return {
        # #    'type': self.type.to_dto(),
        # #    'recipient': self.recipient.to_dto(),
        # #    'mosaicId': self.mosaic_id.to_dto(),
        # #    'fee': self.fee
        # #}

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls()
        # #return cls(
        # #    type=MosaicLevyType.create_from_dto(data['type']),
        # #    recipient=Address.create_from_dto(data['recipient']),
        # #    mosaic_id=MosaicId.create_from_dto(data['mosaicId']),
        # #    fee=data['fee'],
        # #)



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config


@harness.http_test_case({
    'clients': (client.BlockchainHTTP, client.AsyncBlockchainHTTP),
    'tests': [
        {
            #/diagnostic/storage
            'name': 'test_get_diagnostic_storage',
            'params': [],
            'method': 'get_diagnostic_storage',
            'validation': [
                lambda x: (isinstance(x, models.BlockchainStorageInfo), True),
            ]
        },
        {
            #/diagnostic/server
            'name': 'test_get_diagnostic_server',
            'params': [],
            'method': 'get_diagnostic_server',
            'validation': [
                lambda x: (isinstance(x, models.BlockchainServerInfo), True),
            ]
        },
    ],
})
class TestDiagnosticHttp(harness.TestCase):
    pass




from xpxchain import models

gen_hash = '7B631D803F912B00DC0CBED3014BBD17A302BA50B99D233B9C2D9533B842ABDF'

nemesis_signer_public_key = 'B4F12E7C9F6946091E2CB8B6D3A12B50D17CCBBF646386EA27CE2946A7423DCF'
nemesis_harvesting_public_key = '346E56F77F07B19D48B3AEF969EDB01F68A5AC9FAF8E5E007577D71BA66385FB'
api_node_public_key = '460458B98E2BAA36A8E95DE9B320379E89898885B71CF0174E02F1324FAFFAC1'

nemesis = models.Account.create_from_private_key('28FCECEA252231D2C86E1BCF7DD541552BDBBEFBB09324758B3AC199B4AA7B78', models.NetworkType.MIJIN_TEST)
nemesis_signer = models.Account.create_from_private_key('C06B2CC5D7B66900B2493CF68BE10B7AA8690D973B7F0B65D0DAE4F7AA464716', models.NetworkType.MIJIN_TEST)
mosaic_id = models.MosaicId.create_from_hex('0dc67fbe1cad29e3')



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config


@harness.http_test_case({
    'clients': (client.ConfigHTTP, client.AsyncConfigHTTP),
    'tests': [
        {
            'name': 'test_get_config',
            'params': [100,],
            'method': 'get_config',
            'validation': [
                lambda x: (isinstance(x, models.CatapultConfig), True),
            ]
        },
    ],
})
class TestConfigHttp(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config

@harness.http_test_case({
    'clients': (client.NodeHTTP, client.AsyncNodeHTTP),
    'tests': [
        {
            #/node/info
            'name': 'test_get_node_info',
            'params': [],
            'method': 'get_node_info',
            'validation': [
                lambda x: (isinstance(x, models.NodeInfo), True),
                lambda x: (x.public_key, config.api_node_public_key),
            ]
        },
        {
            #/node/time
            'name': 'test_get_node_time',
            'params': [],
            'method': 'get_node_time',
            'validation': [
                lambda x: (isinstance(x, models.NodeTime), True),
            ]
        },
    ],
})
class TestNodeHttp(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from xpxchain import util
from tests import harness
from tests import config
from tests import responses
import os
from binascii import hexlify

class Error(Exception):
    pass


@harness.http_test_case({
    'clients': (client.AccountHTTP, client.AsyncAccountHTTP),
    'tests': [
        {
            #/account/{accountId}
            'name': 'test_get_account_info',
            'params': [config.nemesis.address],
            'method': 'get_account_info',
            'validation': [
                lambda x: (x.public_key, config.nemesis.public_key.upper()),
            ]
        },
        {
            #/account
            'name': 'test_get_accounts_info',
            'params': [[config.nemesis.address]],
            'method': 'get_accounts_info',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].public_key, config.nemesis.public_key.upper()),
            ]
        },
        {
            #/account/{publicKey}/transaction/unconfirmed
            'name': 'test_unconfirmed_transactions',
            'params': [config.nemesis],
            'method': 'unconfirmed_transactions',
            'validation': [
                lambda x: (len(x), 0),
            ],
        },
        {
            #/account/{publicKey}/transaction/partial
            'name': 'test_aggregate_bonded_transactions',
            'params': [config.nemesis],
            'method': 'aggregate_bonded_transactions',
            'validation': [
                lambda x: (len(x), 0),
            ],
        },
    ],
})
class TestAccountHttp(harness.TestCase):
    def __init__(self, task) -> None:
        super().__init__(task)

        if (task == 'test_account_transactions'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, 100000000)

        elif (task == 'test_incoming_transactions'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, 100000000)
        
        elif (task == 'test_outgoing_transactions'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.alice, 100000000)
            self.send_funds(self.alice, self.bob, 10000000)
        
        elif (task == 'test_multisig_account_info'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.multisig = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

            nemesis_to_bob = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.bob.address,
                mosaics=[models.Mosaic(config.mosaic_id, 10000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            nemesis_to_alice = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.alice.address,
                mosaics=[models.Mosaic(config.mosaic_id, 10000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )
            
            nemesis_to_multisig = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.multisig.address,
                mosaics=[models.Mosaic(config.mosaic_id, 10000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            tx = models.AggregateTransaction.create_complete(
                deadline=models.Deadline.create(),
                inner_transactions=[nemesis_to_alice.to_aggregate(config.nemesis), nemesis_to_bob.to_aggregate(config.nemesis), nemesis_to_multisig.to_aggregate(config.nemesis)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            signed_tx = tx.sign_transaction_with_cosignatories(config.nemesis, config. gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.alice)
            self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

            change_to_multisig = models.ModifyMultisigAccountTransaction.create(
                deadline=models.Deadline.create(),
                min_approval_delta=2,
                min_removal_delta=1,
                modifications=[
                    models.MultisigCosignatoryModification.create(self.alice, models.MultisigCosignatoryModificationType.ADD),
                    models.MultisigCosignatoryModification.create(self.bob, models.MultisigCosignatoryModificationType.ADD),
                ],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            tx = models.AggregateTransaction.create_complete(
                deadline=models.Deadline.create(),
                inner_transactions=[change_to_multisig.to_aggregate(self.multisig)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            signed_tx = tx.sign_transaction_with_cosignatories(self.multisig, config.gen_hash, [self.alice, self.bob], fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.multisig)
            self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        elif (task == 'test_multisig_account_graph_info'):
            self.alice = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.bob = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.mike = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.multisig = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.multisig2 = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))

            nemesis_to_bob = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.bob.address,
                mosaics=[models.Mosaic(config.mosaic_id, 100000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            nemesis_to_alice = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.alice.address,
                mosaics=[models.Mosaic(config.mosaic_id, 100000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )
            
            nemesis_to_mike = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.mike.address,
                mosaics=[models.Mosaic(config.mosaic_id, 100000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )
            
            nemesis_to_multisig = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.multisig.address,
                mosaics=[models.Mosaic(config.mosaic_id, 100000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )
            
            nemesis_to_multisig2 = models.TransferTransaction.create(
                deadline=models.Deadline.create(),
                recipient=self.multisig2.address,
                mosaics=[models.Mosaic(config.mosaic_id, 100000000)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            tx = models.AggregateTransaction.create_complete(
                deadline=models.Deadline.create(),
                inner_transactions=[
                    nemesis_to_alice.to_aggregate(config.nemesis), 
                    nemesis_to_bob.to_aggregate(config.nemesis), 
                    nemesis_to_mike.to_aggregate(config.nemesis), 
                    nemesis_to_multisig.to_aggregate(config.nemesis), 
                    nemesis_to_multisig2.to_aggregate(config.nemesis),
                ],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            signed_tx = tx.sign_transaction_with_cosignatories(config.nemesis, config. gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.alice)
            self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

            change_to_multisig = models.ModifyMultisigAccountTransaction.create(
                deadline=models.Deadline.create(),
                min_approval_delta=2,
                min_removal_delta=1,
                modifications=[
                    models.MultisigCosignatoryModification.create(self.alice, models.MultisigCosignatoryModificationType.ADD),
                    models.MultisigCosignatoryModification.create(self.bob, models.MultisigCosignatoryModificationType.ADD),
                ],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            tx = models.AggregateTransaction.create_complete(
                deadline=models.Deadline.create(),
                inner_transactions=[change_to_multisig.to_aggregate(self.multisig)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            signed_tx = tx.sign_transaction_with_cosignatories(self.multisig, config.gen_hash, [self.alice, self.bob], fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.multisig)
            self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

            change_to_multisig = models.ModifyMultisigAccountTransaction.create(
                deadline=models.Deadline.create(),
                min_approval_delta=2,
                min_removal_delta=1,
                modifications=[
                    models.MultisigCosignatoryModification.create(self.mike, models.MultisigCosignatoryModificationType.ADD),
                    models.MultisigCosignatoryModification.create(self.multisig, models.MultisigCosignatoryModificationType.ADD),
                ],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            tx = models.AggregateTransaction.create_complete(
                deadline=models.Deadline.create(),
                inner_transactions=[change_to_multisig.to_aggregate(self.multisig2)],
                network_type=models.NetworkType.MIJIN_TEST,
            )

            signed_tx = tx.sign_transaction_with_cosignatories(self.multisig2, config.gen_hash, [self.alice, self.bob, self.mike], fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.multisig2)
            self.assertEqual(isinstance(tx, models.AggregateTransaction), True)

        elif (task == 'test_account_names'):
            self.mike = models.Account.generate_new_account(models.NetworkType.MIJIN_TEST, entropy = lambda x: os.urandom(32))
            self.send_funds(config.nemesis, self.mike, 1000000000)
            namespace_name = 'foo' + hexlify(os.urandom(4)).decode('utf-8')

            # Create namespace
            tx = models.RegisterNamespaceTransaction.create_root_namespace(
                deadline=models.Deadline.create(),
                network_type=models.NetworkType.MIJIN_TEST,
                namespace_name=namespace_name,
                duration=60
            )

            signed_tx = tx.sign_with(self.mike, config. gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.mike)
            self.assertEqual(isinstance(tx, models.RegisterNamespaceTransaction), True)
            self.assertEqual(tx.namespace_name, namespace_name)

            # Create sub namespace
            tx = models.RegisterNamespaceTransaction.create_sub_namespace(
                deadline=models.Deadline.create(),
                network_type=models.NetworkType.MIJIN_TEST,
                namespace_name='bar',
                parent_namespace=namespace_name
            )

            signed_tx = tx.sign_with(self.mike, config. gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.mike)
            self.assertEqual(isinstance(tx, models.RegisterNamespaceTransaction), True)
            self.assertEqual(tx.namespace_name, 'bar')

            self.mikes_namespace = namespace_name + ".bar"

            # Link address alias
            tx = models.AddressAliasTransaction.create(
                deadline=models.Deadline.create(),
                network_type=models.NetworkType.MIJIN_TEST,
                max_fee=1,
                action_type=models.AliasActionType.LINK,
                namespace_id=models.NamespaceId(self.mikes_namespace),
                address=self.mike.address
            )

            signed_tx = tx.sign_with(self.mike, config. gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

            with client.TransactionHTTP(responses.ENDPOINT) as http:
                http.announce(signed_tx)

            tx = self.listen(self.mike)
            self.assertEqual(isinstance(tx, models.AddressAliasTransaction), True)
            self.assertEqual(tx.action_type, models.AliasActionType.LINK)
            self.assertEqual(tx.address, self.mike.address) 

    async def listen(self, account):
        async with client.Listener(f'{responses.ENDPOINT}/ws') as listener:
            await listener.confirmed(account.address)
            await listener.status(account.address)

            async for m in listener:
                if (m.channel_name == 'status'):
                    raise Error(m.message)
                elif (m.channel_name == 'confirmedAdded'):
                    return m.message

    def send_funds(self, sender, recipient, amount):
        tx = models.TransferTransaction.create(
            deadline=models.Deadline.create(),
            recipient=recipient.address,
            mosaics=[models.Mosaic(config.mosaic_id, amount)],
            network_type=models.NetworkType.MIJIN_TEST,
        )

        signed_tx = tx.sign_with(sender, config.gen_hash, fee_strategy=util.FeeCalculationStrategy.MEDIUM)

        with client.TransactionHTTP(responses.ENDPOINT) as http:
            http.announce(signed_tx)

        tx = self.listen(sender)
        self.assertEqual(isinstance(tx, models.TransferTransaction), True)
        self.assertEqual(tx.recipient, recipient.address)
        self.assertEqual(len(tx.mosaics), 1)
        self.assertEqual(tx.mosaics[0].id, config.mosaic_id)
        self.assertEqual(tx.mosaics[0].amount, amount)

        return tx
    
    def test_multisig_account_info(self):
        with client.AccountHTTP(responses.ENDPOINT) as http:
            info = http.get_multisig_account_info(self.multisig.address)
            self.assertEqual(isinstance(info, models.MultisigAccountInfo), True) 
    
            
    def test_multisig_account_graph_info(self):
        with client.AccountHTTP(responses.ENDPOINT) as http:
            info = http.get_multisig_account_graph_info(self.multisig.address)
            self.assertEqual(isinstance(info, models.MultisigAccountGraphInfo), True) 


    def test_account_names(self):
        with client.AccountHTTP(responses.ENDPOINT) as http:
            info = http.get_account_names([self.mike.address])
            self.assertEqual(len(info), 1)
            self.assertEqual(isinstance(info[0], models.AccountNames), True)
            self.assertEqual(len(info[0].names), 1)
            self.assertEqual(info[0].names[0], models.NamespaceId(self.mikes_namespace))
    
    
    def test_account_transactions(self):
        with client.AccountHTTP(responses.ENDPOINT) as http:
            info = http.transactions(self.alice)
            self.assertEqual(len(info), 1)
            self.assertEqual(len(info[0].mosaics), 1)
            self.assertEqual(info[0].mosaics[0], models.Mosaic(config.mosaic_id, 100000000))
            self.assertEqual(info[0].signer.public_key, config.nemesis.public_key.upper())
    
    
    def test_incoming_transactions(self):
        with client.AccountHTTP(responses.ENDPOINT) as http:
            info = http.incoming_transactions(self.alice)
            self.assertEqual(len(info), 1)
            self.assertEqual(len(info[0].mosaics), 1)
            self.assertEqual(info[0].mosaics[0], models.Mosaic(config.mosaic_id, 100000000))
            self.assertEqual(info[0].signer.public_key, config.nemesis.public_key.upper())
    
    
    def test_outgoing_transactions(self):
        with client.AccountHTTP(responses.ENDPOINT) as http:
            info = http.outgoing_transactions(self.alice)
            self.assertEqual(len(info), 1)
            self.assertEqual(len(info[0].mosaics), 1)
            self.assertEqual(info[0].mosaics[0], models.Mosaic(config.mosaic_id, 10000000))
            self.assertEqual(info[0].recipient, self.bob.address)



from xpxchain import client
from xpxchain import models
from tests import harness


@harness.http_test_case({
    'clients': (client.NetworkHTTP, client.AsyncNetworkHTTP),
    'tests': [
        {
            #/network
            'name': 'test_get_network_type',
            'params': [],
            'method': 'get_network_type',
            'validation': [
                lambda x: (isinstance(x, models.NetworkType), True),
                lambda x: (x, models.NetworkType.MIJIN_TEST),
            ]
        },
    ],
})
class TestNetworkHttp(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness


@harness.http_test_case({
    'clients': (client.MosaicHTTP, client.AsyncMosaicHTTP),
    'tests': [
        {
            #/mosaic/{mosaicId}
            'name': 'test_get_mosaic',
            'params': [models.MosaicId.create_from_hex('0dc67fbe1cad29e3')],
            'method': 'get_mosaic',
            'validation': [
                lambda x: (x.owner.public_key, 'B4F12E7C9F6946091E2CB8B6D3A12B50D17CCBBF646386EA27CE2946A7423DCF'),
            ]
        },
        {
            #/mosaic
            'name': 'test_get_mosaics',
            'params': [[models.MosaicId.create_from_hex('0dc67fbe1cad29e3')]],
            'method': 'get_mosaics',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].owner.public_key, 'B4F12E7C9F6946091E2CB8B6D3A12B50D17CCBBF646386EA27CE2946A7423DCF'),
            ]
        },
        {
            #/mosaic/names
            'name': 'test_get_mosaic_names',
            'params': [[models.MosaicId.create_from_hex('0dc67fbe1cad29e3')]],
            'method': 'get_mosaic_names',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].names[0], 'prx.xpx'),
            ]
        },
    ],
})
class TestMosaicHttp(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config

@harness.http_test_case({
    'clients': (client.NamespaceHTTP, client.AsyncNamespaceHTTP),
    'tests': [
        {
            #/namespace/{namespaceId}
            'name': 'test_get_namespace',
            'params': [models.NamespaceId('prx.xpx')],
            'method': 'get_namespace',
            'validation': [
                lambda x: (isinstance(x, models.NamespaceInfo), True),
                lambda x: (x.owner.public_key, config.nemesis_signer_public_key),
            ]
        },
        {
            #/account/{accountId}/namespaces
            'name': 'test_get_namespaces_from_account',
            'params': [config.nemesis_signer.address],
            'method': 'get_namespaces_from_account',
            'validation': [
                lambda x: (len(x), 4),
                lambda x: (isinstance(x[0], models.NamespaceInfo), True),
            ]
        },
        {
            #/account/namespaces
            'name': 'test_get_namespaces_from_accounts',
            'params': [[config.nemesis_signer.address]],
            'method': 'get_namespaces_from_accounts',
            'validation': [
                lambda x: (len(x), 4),
                lambda x: (isinstance(x[0], models.NamespaceInfo), True),
            ]
        },
        {
            #/namespace/names
            'name': 'test_get_namespaces_name',
            'params': [[models.NamespaceId.create_from_hex('b16d77fd8b6fb3be')]],
            'method': 'get_namespaces_name',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].name, 'prx'),
            ]
        },
    ],
})
class TestNamespaceHttp(harness.TestCase):
    pass



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import config
from tests import responses

@harness.http_test_case({
    'clients': (client.BlockchainHTTP, client.AsyncBlockchainHTTP),
    'tests': [
        {
            #/blocks/{height}/limit/{limit}
            'name': 'test_get_blocks_by_height_with_limit',
            'params': [25, 25],
            'method': 'get_blocks_by_height_with_limit',
            'validation': [
                lambda x: (len(x), 25),
                lambda x: (isinstance(x[0], models.BlockInfo), True)
            ]
        },
        {
            #/block/{height}
            'name': 'test_get_block_by_height',
            'params': [25],
            'method': 'get_block_by_height',
            'validation': [
                lambda x: (isinstance(x, models.BlockInfo), True)
            ]
        },
        {
            #/block/{height}/transactions
            'name': 'test_get_block_transactions',
            'params': [1],
            'method': 'get_block_transactions',
            'validation': [
                lambda x: (len(x) > 0, True),
                lambda x: (isinstance(x[0], models.Transaction), True),
                lambda x: (x[0].signer.public_key, config.nemesis_signer_public_key),
            ]
        },
        {
            #/block/{height}/receipts
            'name': 'test_get_block_receipts',
            'params': [25],
            'method': 'get_block_receipts',
            'validation': [
                lambda x: (isinstance(x, models.Statements), True),
                lambda x: (len(x.transaction_statements), 1),
                lambda x: (isinstance(x.transaction_statements[0], models.TransactionStatement), True),
                lambda x: (len(x.transaction_statements[0].receipts), 1),
                lambda x: (isinstance(x.transaction_statements[0].receipts[0], models.BalanceChangeReceipt), True),
                lambda x: (x.transaction_statements[0].receipts[0].account, config.nemesis_harvesting_public_key)
            ]
        },
        #TODO: Looks like there's no hash in receipts yet 
        #/block/{height}/receipt/{hash}/merkle
    ],
})
class TestBlockHttp(harness.TestCase):
    def __init__(self, task) -> None:
        super().__init__(task)

        if (task == 'test_get_merkle_by_hash_in_block'):
            with client.BlockchainHTTP(responses.ENDPOINT) as http:
                reply = http.get_block_transactions(1)
                self.tx_hash = reply[0].transaction_info.hash

    
    def test_get_merkle_by_hash_in_block(self):
        with client.BlockchainHTTP(responses.ENDPOINT) as http:
            reply = http.get_merkle_by_hash_in_block(1, self.tx_hash)    
            self.assertEqual(isinstance(reply, models.MerkleProofInfo), True)
            self.assertEqual(len(reply.merkle_path) > 0, True)
            self.assertEqual(isinstance(reply.merkle_path[0], models.MerklePathItem), True)



"""
    mosaic_name
    ===========

    Describes a mosaic by name and identifier.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

import typing
from .mosaic_id import MosaicId
from ..blockchain.network_type import OptionalNetworkType
from ..namespace.namespace_id import NamespaceId
from ... import util

__all__ = ['MosaicName']


@util.inherit_doc
@util.dataclass(frozen=True)
class MosaicName(util.DTO):
    """
    Mosaic name and identifiers.

    :param mosaic_id: Mosaic ID.
    :param names: Mosaic name.

    DTO Format:
        .. code-block:: yaml

            MosaicNameDTO:
                mosaicId: UInt64DTO
                names: string
    """

    mosaic_id: MosaicId
    names: typing.Sequence[str]

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'mosaicId', 'names'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'mosaicId': util.u64_to_dto(int(self.mosaic_id)),
            'names': self.names,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            mosaic_id=MosaicId(util.u64_from_dto(data['mosaicId'])),
            names=data['names'],
        )
